/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <string.h>
#include <gnome.h>
#include "pan.h"



void
article_add_flag (article_data *adata, int flag)
{
	adata->state |= flag;
	article_update (adata);
}


void
article_remove_flag (article_data *adata, int flag)
{
	adata->state &= ~flag;
	article_update (adata);
}


/*---[ article_copy ]-------------------------------------------------
 * like a strdup function, returns an exact copy of the article passed
 * into it as a parameter
 *--------------------------------------------------------------------*/
article_data *
article_copy (article_data *src)
{
	article_data *dest;

	dest = (article_data *) g_malloc0 (sizeof(article_data));

	dest->number = g_strdup (src->number);
	dest->subject = g_strdup (src->subject);
	dest->author = g_strdup (src->author);
	dest->message_id = g_strdup (src->message_id);
	dest->references = g_strdup (src->references);
	dest->date = src->date;
	dest->linecount = src->linecount;
	dest->parts = src->parts;
	dest->part = src->part;
	dest->group = src->group;
	dest->header = g_strdup (src->header);
	dest->body = g_strdup (src->body);

	return dest;
}


/*--------------------------------------------------------------------
 * removes article from the database for that group, also removes it
 * from the displayed articlelist
 *
 * FIXME: If it is a parent node, it will cause the thread to disappear
 *--------------------------------------------------------------------*/
void
article_delete (void)
{
	GList *list, *next;
	article_data *adata;
	group_data *gdata;

	gtk_clist_freeze (GTK_CLIST(Pan.article_ctree));
	next = NULL;
	list = GTK_CLIST (Pan.article_ctree)->selection; 
	while (list) {
		next = list->next;
		adata = (article_data *) gtk_ctree_node_get_row_data(GTK_CTREE(Pan.article_ctree), list->data);		
		if (adata) {
			db_erase_article (adata);
			gdata = adata->group;
			gdata->articlelist = g_slist_remove(gdata->articlelist, adata);
			article_free (adata);
		}
		gtk_ctree_remove_node (GTK_CTREE(Pan.article_ctree), list->data);
		list = next;
	}
	gtk_clist_thaw (GTK_CLIST(Pan.article_ctree));
}


/*--------------------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
void
article_delete_body (article_data *adata)
{
	if (adata->body)
		g_free (adata->body);

	adata->state &= ~STATE_DECODED;
}


int
article_find (article_data *adata, gchar *message_id)
{
	if (adata == NULL)
		return 1;

	return (strcmp(adata->message_id, message_id));
}


void
article_free (article_data *adata)
{
	g_free (adata->number);
	g_free (adata->subject);
	g_free (adata->author);
	g_free (adata->message_id);
	g_free (adata->references);
	g_free (adata->header);
	g_free (adata->body);
	g_free (adata);
}


/*---[ article_get_current ]------------------------------------------
 * returns the current displayed (or selected) article..
 *--------------------------------------------------------------------*/
article_data *
article_get_current (void)
{
	/* FIXME: Return something */
	return NULL;
}


article_data *
article_in_memory (char *message_id)
{
	GtkWidget *node;

	GTK_CTREE_NODE(node) =
		gtk_ctree_find_by_row_data_custom (GTK_CTREE(Pan.article_ctree),
						   NULL, message_id,
						   (GCompareFunc) article_find);
	if (node)
		return (GTK_CTREE_ROW(node)->row.data);
	else
		return NULL;
}


/*---[ article_mark_read ]--------------------------------------------
 * mark an article as read
 *--------------------------------------------------------------------*/
void
article_mark_read (article_data *adata)
{
	if (adata)
		adata->state |= STATE_READ;
}


/*---[ article_mark_unread ]------------------------------------------
 * mark an article as unread
 *--------------------------------------------------------------------*/
void
article_mark_unread (article_data *adata)
{
	if (adata)
		adata->state &= ~STATE_READ;
}


/*--------------------------------------------------------------------
 * should only be called from the network thread
 *--------------------------------------------------------------------*/
void
article_update (article_data *adata)
{
	article_data *p;

	g_assert (adata && adata->message_id);

	db_store_article (adata->group, adata);

	pan_lock ();
	p = article_in_memory (adata->message_id);
	/* Make sure relevant variables are in sync */
	if (p) {
		p->state = adata->state;
		articlelist_update_node (p);
	}
	pan_unlock ();
}


/*
 * Extracts the requested header from adata->header;
 * hdr is the case sensitive name, without the colon
 * Doesn't handle wrapped lines
 */
gchar *
extract_header (article_data *adata, gchar *hdr)
{
	gchar *start, *end, *header;

	header = g_strconcat (hdr, ": ", NULL);
	start = strstr (adata->header, header);
	g_free (header);
	if (start == NULL)
		return NULL;
	else {
		start = strchr (start, ':');
		end = strchr (start, '\n');
		return g_strndup (start + 2, end - start - 3);
	}
}
