/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for mo0re details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <gnome.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include "pan.h"

GSList *queue = NULL;
GSList *queue_last = NULL;

pthread_mutex_t qlock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t qcond_m = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t qcond = PTHREAD_COND_INITIALIZER;


int
queue_active (int set)
{
	static int qactive;
	switch (set) {
	case 0:
		qactive = 0;
		break;
	case 1:
		qactive = 1;
		break;
	}
	return(qactive);
}


void
queue_add_article_download (article_data *adata, guint flags)
{
	queue_data *qdata;
	GSList *temp_list;

	qdata = (queue_data *) g_malloc0(sizeof(queue_data));
	qdata->gdata = adata->group;
	qdata->ptr = (char *)g_strdup(adata->message_id);
	qdata->flags = flags;
	temp_list = g_slist_alloc();
	temp_list->data = qdata;
	pthread_mutex_lock(&qlock);
	if (flags & QUEUE_FOREGROUND)
		queue_prepend(qdata);
	else if (flags & QUEUE_BACKGROUND)
		queue_append(qdata);
	else
		g_error("Unreachable condition reached\n");
	pthread_mutex_unlock(&qlock);
	/* Signal the addition */
	pthread_mutex_lock(&qcond_m);
	pthread_cond_broadcast(&qcond);
	pthread_mutex_unlock(&qcond_m);
}


void
queue_add_articlelist_download (group_data *gdata, guint flags)
{
	queue_data *qdata;

	qdata = (queue_data *) g_malloc0(sizeof(queue_data));
	qdata->gdata = gdata;
	qdata->flags = flags;
	pthread_mutex_lock (&qlock);
	if (flags & QUEUE_FOREGROUND)
		queue_prepend (qdata);
	else if (flags & QUEUE_BACKGROUND)
		queue_append (qdata);
	else
		g_error("Unreachable condition reached\n");

	pthread_mutex_unlock(&qlock);
	/* Signal the addition */
	pthread_mutex_lock(&qcond_m);
	pthread_cond_broadcast(&qcond);
	pthread_mutex_unlock(&qcond_m);
}


void
queue_add_grouplist_download (void)
{
	queue_data *qdata;

	qdata = (queue_data *) g_malloc0(sizeof(queue_data));
	qdata->flags = QUEUE_GROUPLIST_DOWNLOAD;
	pthread_mutex_lock(&qlock);
	queue = g_slist_prepend(queue, qdata);
	pthread_mutex_unlock(&qlock);
	/* Signal the addition */
	pthread_mutex_lock(&qcond_m);
	pthread_cond_broadcast(&qcond);
	pthread_mutex_unlock(&qcond_m);
}


void
queue_add_article_decode (article_data *adata, guint flags)
{
	queue_data *qdata;
	GSList *l;

	g_assert (adata && flags);

	qdata = (queue_data *) g_malloc0(sizeof(queue_data));
	qdata->gdata = adata->group;

	qdata->ptr = g_slist_append (qdata->ptr, g_strdup(adata->message_id));
	for (l = adata->threads; l; l = l->next) {
		qdata->ptr = g_slist_append (qdata->ptr, g_strdup(((article_data *)l->data)->message_id));
	}
	qdata->flags = flags;
	pthread_mutex_lock (&qlock);
	queue_append (qdata);
	pthread_mutex_unlock (&qlock);
	/* Signal the addition */
	pthread_mutex_lock (&qcond_m);
	pthread_cond_broadcast (&qcond);
	pthread_mutex_unlock (&qcond_m);
}


void
queue_add_post (gchar *message)
{
	queue_data *qdata;

	qdata = (queue_data *) g_malloc0(sizeof(queue_data));
	qdata->ptr = (char *)message;
	qdata->flags = QUEUE_ARTICLE_POST;
	pthread_mutex_lock(&qlock);
	queue_prepend(qdata);
	pthread_mutex_unlock(&qlock);

	/* Signal the addition */
	pthread_mutex_lock(&qcond_m);
	pthread_cond_broadcast(&qcond);
	pthread_mutex_unlock(&qcond_m);
}


void
queue_append (queue_data *qdata)
{
	GSList *l;

	l = g_slist_alloc();
	l->data = qdata;

	if (queue) {
		g_assert(queue_last);
		g_slist_concat(queue_last, l);
		queue_last = l;
	} else
		queue = queue_last = l;
}


/*---[ queue_get_next ]-----------------------------------------------
 * return next item on queue, remove link to current one
 *--------------------------------------------------------------------*/
queue_data*
queue_get_next (void)
{
	queue_data *qdata;

	pthread_mutex_lock(&qlock);
	if (queue) {
		qdata = (queue_data *) queue->data;
		queue = g_slist_remove_link (queue, queue);
		if (queue == NULL)
			queue_last = NULL;
		pthread_mutex_unlock (&qlock);
		return qdata;
	}
	else {
		pthread_mutex_unlock (&qlock);
		return NULL;
	}
}


void
queue_init (void)
{
	/* Fire up the queue thread */
	pthread_create (&Pan.thread, NULL, (void *) &queue_processing, NULL);
	pthread_detach (Pan.thread);
}


/*--------------------------------------------------------------------
 * put an item at the front of the queue
 *--------------------------------------------------------------------*/
void
queue_prepend (queue_data *qdata)
{
	if (queue_last)
		queue = g_slist_prepend (queue, qdata);
	else
		queue_last = queue = g_slist_prepend (queue, qdata);
}


void
queue_processing (void)
{
	struct timeval now;
	struct timespec timeout;
	int retcode;
	queue_data *qdata;
	time_t last_run, current_time;
	
	while (1) {
	
		pthread_mutex_lock (&qcond_m);
		gettimeofday (&now, NULL);
		timeout.tv_sec = now.tv_sec + 20;
		timeout.tv_nsec = now.tv_usec * 1000;
		retcode = 0;
	
		/* Wait for something to hit the queue */
		retcode = pthread_cond_timedwait (&qcond, &qcond_m, &timeout);
		if ((retcode == ETIMEDOUT) && (queue == NULL)) {
			/* Nothing hit the queue before timeout */
			if (is_online()) {
				time (&current_time);
				if ((current_time - last_run) > 300)
					/* We've been idle for 300 seconds, so go offline */
					go_offline ();
				else
					/* Send HELP as a no-operation to keep the connection alive */
					nntp_send_noop ();
			}
		} else {
			/* Something hit the queue, do it */
			pthread_mutex_unlock(&qcond_m);
			queue_active(1);
			while ((qdata = queue_get_next())) {
				queue_run (qdata);
			}
			/* Record the last time we ran queue */
			queue_active(0);
			pthread_mutex_lock(&qcond_m);
			time(&last_run);
		}
		pthread_mutex_unlock (&qcond_m);
	}
}


/*--------------------------------------------------------------------
 * remove all items from the queue
 *--------------------------------------------------------------------*/
void
queue_purge (void)
{
	GSList *p;

	pthread_mutex_lock (&qlock);
	for (p = queue; p; p = p->next) {
		/* FIXME: Remove the "queued" icon in the article_ctree */
		g_free (p->data);
	}
	g_slist_free (queue);
	queue = NULL;
	queue_last = NULL;

	pthread_mutex_unlock (&qlock);
}

/*
void
stop_queue_processing (void)
{
	pthread_mutex_lock(&qlock);
	g_slist_free(queue);
	queue = NULL;
	queue_last = NULL;
	q_active = FALSE;
	pthread_mutex_unlock(&qlock);
	go_offline();
}
*/

int
queue_connect_wait (void)
{
	while (1) {
		//pthread_mutex_lock(&online_m);
		if (online == TRUE) {
			//pthread_mutex_unlock(&online_m);
			return TRUE;
		}
		//pthread_mutex_unlock(&online_m);
		if (stop_all || must_exit) {
			return FALSE;
		}
		usleep (1000);
	}		
}

void
queue_run_grouplist (queue_data *qdata)
{
	gint count;

	/* Free the old one if it exists */
	grouplist_free (Pan.grouplist);
	Pan.grouplist = NULL;

	/* Download the new one */
	go_online();
	if (queue_connect_wait() == FALSE)
		return;
	if ((count = nntp_grouplist_download()) >= 0) {
		/* Sort it and show it */
		gui_set_status_va (_("Sorting %d groups."), count);
		Pan.grouplist = g_slist_sort (Pan.grouplist, (GCompareFunc)compare_group_data);

		grouplist_save ();
		grouplist_update ();
	}
}


void
queue_run_article (queue_data *qdata)
{
	article_data *adata;
	
	if ((adata = db_fetch_article(qdata->gdata, qdata->ptr)) == NULL) {
//		gui_set_info_va (_("Couldn't find article header for message_id: %s"), qdata->ptr);
		log_add_va (_("Couldn't find article header for message_id '%s'"), qdata->ptr);
		g_free (qdata->ptr);
		return;
	}
	g_free (qdata->ptr);
	
	article_remove_flag (adata, STATE_DOWNLOAD_QUEUED);

	/* Clear the one already displayed, if any */
	if (qdata->flags & QUEUE_FOREGROUND) {
		text_clear();
	}

	/* Check the cache before we download */
	if (acache_hit_header(adata) || acache_hit_body(adata)) {
		go_online();
		if (queue_connect_wait() == FALSE)
			return;
		if (nntp_article_download(qdata->gdata, adata) != 0) {
			article_free (adata);
			return;
		}
	} 
	
	/* Display, et cetera */
	if (qdata->flags & QUEUE_FOREGROUND)
		text_update ((article_data *)adata);

	if ((adata->state == STATE_NONE) || (adata->state == STATE_NEW))
		group_add_attrib_i(qdata->gdata, "Unread", -1);
	article_add_flag(adata, STATE_READ);

	article_free(adata);
}


void
queue_run_decode (queue_data *qdata)
{
	article_data *adata;

	/* Make sure the first one is in the db */
	if ((adata = db_fetch_article (qdata->gdata, ((GSList *)qdata->ptr)->data)) == NULL) {
//		gui_set_info_va (_("Couldn't find article header for message_id: %s"), ((GSList *)qdata->ptr)->data);
		log_add_va (_("Couldn't find article header for message_id '%s'"), ((GSList *)qdata->ptr)->data);
		return;
	}
	
	/* Go online and get them */
	go_online();
	if (queue_connect_wait() == FALSE)
		return;
	if (nntp_article_decode (qdata->gdata, adata, qdata->ptr) == 0) {
		/* Decode them */
		gui_set_status_va (_("Preparing article '%s' for decode..."), adata->subject);
		gui_set_progress (0);
		if ((adata->state == STATE_NONE) || (adata->state == STATE_NEW))
			group_add_attrib_i (qdata->gdata, "unread", -1);
		if (article_decode (adata, (GSList *)qdata->ptr) == -1) {
			/* Decode failed */
			adata->state &= ~STATE_DECODE_QUEUED;
			article_add_flag (adata, STATE_DECODE_FAILED);
		} else {
			/* Decode succeeded */
			adata->state &= ~STATE_DECODE_QUEUED;
			article_add_flag (adata, STATE_DECODED);
		}
	}

	article_free (adata);
}


void
queue_run_articlelist (queue_data *qdata)
{
	int count = 0;
	group_data *gdata;

	/* Just to make things more readable */
	gdata = qdata->gdata;
	
	go_online();
	if (queue_connect_wait() == FALSE)
		return;

	if ((count = nntp_articlelist_download(gdata, qdata->flags)) > 0) {
//		gui_set_info_va (_("Retrieved %d article headers"), count);
		log_add_va (_("Retrieved %d article headers from %s"), count, gdata->name);

		/* Now we thread the articles */
		gui_set_status (_("Threading article list"));
		gui_set_progress (0);
		gdata->articlelist = thread_list_by_multipart (gdata->articlelist, gdata);
		if (must_exit) {
			go_offline ();
			return;
		}
		
		gdata->articlelist = thread_list_by_subject (gdata->articlelist, gdata);
		if (must_exit) {
			go_offline ();
			return;
		}
		
		gdata->articlelist = thread_list_by_reference (gdata->articlelist, gdata);
		if (must_exit) {
			go_offline ();
			return;
		}
		
		gui_set_status (_("Sorting article headers"));
		articlelist_sort ((group_data *) gdata);
		
		grouplist_update_row (gdata);
	
		if (qdata->flags & QUEUE_FOREGROUND) {
			gui_set_status(_("Displaying article headers"));
			articlelist_update((group_data *) gdata);
		}
	}
	else if (count == 0) {
//		gui_set_info(_("No new articles"));
		log_add_va (_("No new articles for group %s"), gdata->name);
	}
	else {
		/* Error */
	}
}


void
queue_run_post (queue_data *qdata)
{
	go_online();
	if (queue_connect_wait() == FALSE)
		return;
	if (nntp_post((char *)qdata->ptr) == 0) {
//		gui_set_info (_("Message posted."));
		log_add (_("Message posted.")); /* FIXME: message '<subject here>' posted */
	}
}


void
queue_run (queue_data *qdata)
{
	g_assert(qdata);
	
	switch (qdata->flags & 0x003F) {
	case QUEUE_ARTICLE_DOWNLOAD:
		queue_run_article (qdata);
		break;
	case QUEUE_ARTICLELIST_LOAD:
		articlelist_load (qdata);
		break;
	case QUEUE_ARTICLELIST_DOWNLOAD_ALL:
	case QUEUE_ARTICLELIST_DOWNLOAD_NEW:
	case QUEUE_ARTICLELIST_DOWNLOAD_SAMPLE:
		queue_run_articlelist (qdata);
		break;
	case QUEUE_GROUPLIST_DOWNLOAD:
		queue_run_grouplist (qdata);
	break;
	case QUEUE_ARTICLE_POST:
		queue_run_post (qdata);
		break;
	case QUEUE_ARTICLE_DECODE:
		queue_run_decode (qdata);	
		break;
	default:
		g_error("Reached default in queue_processing()\n");
		break;
	}
	/* Put stuff here you want to do after processing
	 * queue items */
	g_free (qdata);
	gui_clean ();
}
