/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <errno.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <strings.h>
#include <gnome.h>
#include <pthread.h>
#include <sys/types.h>
#include <fcntl.h>
#include "pan.h"

extern int h_errno;

int sockfd;
int bufpos = 0;
int available = 0;
char inbuf[1024];


void
tcp_close (void)
{
	close (sockfd);
	bufpos = 0;
	available = 0;
}


int
tcp_open (void)
{
	struct sockaddr_in serv_addr;
	struct servent serv_info;
	struct hostent host_info;
	int on;
	unsigned long inaddr;
	struct servent *sp;
	struct hostent *hp;
#ifdef TRY_AGAIN
	unsigned int max_retries = 3;
#endif

	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;

	/* Get service information */
	if ((sp = getservbyport(htons(atoi(Pan.active_server->port)), "tcp")) == NULL) {
		return (-1);
	}
	serv_info = *sp;
	serv_addr.sin_port = sp->s_port;

	/* Try to convert host name as dotted decimal */
	if ((inaddr = inet_addr(Pan.active_server->address)) != -1) {
		bcopy((char *) &inaddr, (char *) &serv_addr.sin_addr,
			  sizeof(inaddr));
		host_info.h_name = NULL;
	} else {					/* If that failed, then look up the host name */
		while (NULL == (hp = gethostbyname(Pan.active_server->address))) {
#ifdef TRY_AGAIN
			max_retries--;
			if (max_retries && (h_errno == TRY_AGAIN)) {
				sleep(1);
				continue;
			}
#endif
			return (-1);
		}
		host_info = *hp;
		bcopy(hp->h_addr, (char *) &serv_addr.sin_addr, hp->h_length);
	}

	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		return (-1);
	}
	if (connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr))
		< 0) {
		close(sockfd);
		return (-1);
	}
	on = 1;
	if (setsockopt
		(sockfd, SOL_SOCKET, SO_KEEPALIVE, (char *) &on, sizeof(on)) < 0) {
		close(sockfd);
		return (-1);
	}
	return (0);
}

/* simple read buffering to hopefully speed things up. */
int 
read_char (char *c)
{
	int retval = 0;
	fd_set rfds;
	struct timeval tv;

	if (bufpos >= available) {
		FD_ZERO(&rfds);
		FD_SET(sockfd, &rfds);
		/* We will timeout after 60 seconds. */
		tv.tv_sec = 60;
		tv.tv_usec = 0;
		retval = select(sockfd + 1, &rfds, NULL, NULL, &tv);
		if (retval <= 0) {
			g_warning("Timed out waiting for select()\n");
			return -1;
		}
		available = read(sockfd, inbuf, 1024);
		bufpos = 0;
		if (available <= 0)
			return available;
	}
	*c = inbuf[bufpos];
	bufpos++;
	return 1;
}


/*
 * get_server - read from a socket until \n
 */
char *
get_server (void)
{
	static char *line_buffer = NULL;
	static int maxline = 1;
	char c;
	int i;
	int done = 0;

	if (line_buffer == NULL)
		line_buffer = (char *) g_malloc(maxline);

	for (i = 0; !done; i++) {
		if (read_char(&c) != 1) {
			return NULL;
		}
		if (i >= maxline - 1) {	/* Grow Line Buffer */
			maxline *= 2;
			line_buffer = (char *) g_realloc(line_buffer, maxline);
		}
		line_buffer[i] = c;
		if (c == '\n')
			done++;
	}
	line_buffer[i] = '\0';
	if (i == 0 || line_buffer[i - 1] != '\n') {
		return NULL;
	}
	return line_buffer;
}


int 
put_server (char *line)
{
	int nleft, nwritten, nbytes;

	nbytes = strlen(line);
	nleft = nbytes;
	while (nleft > 0) {
		nwritten = write(sockfd, line, nleft);
		if (nwritten == -1)
			return -1;
		if (nwritten > 0)
			nleft -= nwritten;
		line += nwritten;
	}
	return (nbytes - nleft);
}


int 
put_server_va (const char *format, ...)
{
	va_list args;
	char *line, *ptr;
	int nleft, nwritten, nbytes;

	g_return_val_if_fail(format != NULL, -1);

	va_start(args, format);
	ptr = g_strdup_vprintf(format, args);
	va_end(args);

	line = ptr;
	nbytes = strlen(line);
	nleft = nbytes;
	while (nleft > 0) {
		nwritten = write(sockfd, line, nleft);
		if (nwritten == -1) {
			g_free(ptr);
			return -1;
		}
		if (nwritten > 0)
			nleft -= nwritten;
		line += nwritten;
	}
	return (nbytes - nleft);
}
